/*-------------------------------------------------------------------------+
| Copyright 2020 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view;

import static java.util.stream.Collectors.joining;
import static org.fortiss.tooling.base.ui.annotation.view.AnnotationViewUtils.SELECTED_ANNOTATION_BACKGROUND_COLOR;
import static org.fortiss.tooling.base.ui.annotation.view.AnnotationViewUtils.getBackgroundColorForEntry;
import static org.fortiss.tooling.base.ui.annotation.view.AnnotationViewUtils.getValueLabel;

import java.util.Collection;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.annotation.IAnnotationValueService;
import org.fortiss.tooling.base.annotation.valueprovider.EStructuralFeatureDescriptor;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTreeTableUIProviderBase;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;

import javafx.scene.paint.Color;

/**
 * UIProvider for the Annotation TreeTable.
 */
/* package */ final class AnnotationTreeTableUIProvider
		extends DynamicTreeTableUIProviderBase<AnnotationEntry> {

	/** The {@link AnnotationViewController} this UI provider was created for. */
	private final AnnotationViewController viewController;

	/**
	 * Constructor.
	 */
	AnnotationTreeTableUIProvider(AnnotationViewController annotationViewFXController) {
		this.viewController = annotationViewFXController;
	}

	/** {@inheritDoc} */
	@Override
	public String getLabel(AnnotationEntry entry, int column) {
		IModelElement element = entry.getModelElement();

		switch(column) {
			case 0:
				return this.viewController.getNameForAnnotationEntry(entry);
			case 1:
				if(element instanceof INamedCommentedElement) {
					return ((INamedCommentedElement)element).getComment();
				}
				return "";
			default:
				Class<? extends IAnnotatedSpecification> columnSpecType =
						this.viewController.colIdxAnnotationMap.get(column)
								.getAnnotatedSpecificationType();

				IAnnotationValueProvider<IAnnotatedSpecification> valueProvider =
						entry.getAnnotationValueProvider(columnSpecType);
				EStructuralFeatureDescriptor eStructuralFeatureDescriptor = valueProvider != null
						? valueProvider.getEStructuralFeatureDescriptor() : null;
				Object specificationValue = entry.getSpecificationValue(columnSpecType);
				if(specificationValue instanceof Collection) {
					Collection<?> values = (Collection<?>)specificationValue;
					return "[" + values.stream().map(o -> getLabel(o, eStructuralFeatureDescriptor))
							.collect(joining(", ")) + "]";
				} else if(specificationValue != null) {
					return getLabel(specificationValue, eStructuralFeatureDescriptor);
				}

				return "";
		}

	}

	/** Returns a label for the given value. */
	private String getLabel(Object value,
			EStructuralFeatureDescriptor eStructuralFeatureDescriptor) {
		EObject root = null;
		if(value instanceof IModelElement) {
			root = eStructuralFeatureDescriptor.getValueRootElement((IModelElement)value);
		}
		return getValueLabel(value, root);
	}

	/** {@inheritDoc} */
	@Override
	public Color getBackgroundColor(int column, AnnotationEntry element) {
		if(element.getModelElement().equals(this.viewController.selected)) {
			return SELECTED_ANNOTATION_BACKGROUND_COLOR;
		}

		if(column > 1 && this.viewController.colIdxAnnotationMap.containsKey(column)) {
			Class<IAnnotatedSpecification> specType = this.viewController.colIdxAnnotationMap
					.get(column).getAnnotatedSpecificationType();
			return getBackgroundColorForEntry(element, specType);
		}

		return super.getBackgroundColor(column, element);
	}

	/** {@inheritDoc} */
	@Override
	public void updateValue(AnnotationEntry element, int column, Object value) {
		IAnnotatedSpecification spec = getAnnotation(column, element);
		this.viewController.setAnnotationEntryValue(spec, element, value);

		super.updateValue(element, column, value);
	}

	/** {@inheritDoc} */
	@Override
	public boolean isEditable(int column, AnnotationEntry ae) {
		IAnnotatedSpecification spec = getAnnotation(column, ae);

		// Check 'editable' predicate for specification/annotation contained by model element
		// associated to given annotation entry (if it exists)
		IAnnotationValueProvider<IAnnotatedSpecification> valueProvider =
				spec != null ? ae.getAnnotationValueProvider(spec.getClass()) : null;
		return valueProvider != null && valueProvider.canEdit(spec);
	}

	/** Get annotation specification associated to given annotation entry and column. */
	private IAnnotatedSpecification getAnnotation(int column, AnnotationEntry ae) {

		if(column > 1 && this.viewController.colIdxAnnotationMap.containsKey(column)) {
			ColumnHandle<IAnnotatedSpecification> columnHandle =
					this.viewController.colIdxAnnotationMap.get(column);

			Class<IAnnotatedSpecification> specType = columnHandle.getAnnotatedSpecificationType();
			IAnnotationValueService as = IAnnotationValueService.getInstance();
			IAnnotatedSpecification spec =
					as.getAnnotationEntry(ae.getModelElement()).getSpecification(specType);
			return spec;
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isEditable(int column) {

		// The first two columns are not supposed to be editable.
		if(column <= 1) {
			return false;
		}

		for(AnnotationEntry ae : this.viewController.root.elements) {
			// If any cell is editable in this column, the column is editable.
			if(isEditable(column, ae)) {
				return true;
			}
		}

		return false;
	}
}
