/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH.
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *     Alexander Diewald (fortiss GmbH) - backport to AF3 & Integration
 *                                        improvements
 *
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.control.treetableview;

import static java.lang.Integer.MAX_VALUE;
import static java.util.stream.Collectors.toList;
import static javafx.scene.control.cell.CheckBoxTreeTableCell.forTreeTableColumn;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.stream.Stream;

import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.ObservableList;
import javafx.scene.control.Label;
import javafx.scene.control.SelectionMode;
import javafx.scene.control.TreeItem;
import javafx.scene.control.TreeTableColumn;
import javafx.scene.control.TreeTableRow;
import javafx.scene.control.TreeTableView;
import javafx.scene.control.TreeTableView.TreeTableViewSelectionModel;
import javafx.scene.input.TransferMode;
import javafx.util.Callback;

/**
 * A JavaFX {@link TreeTableView} based on an {@link DynamicTreeContentProviderBase} and
 * {@link DynamicTreeTableUIProviderBase}. This class is intended to be used as a wrapper for
 * {@link TreeTableView} and its rather complex expert API.
 * <P>
 * This class currently supports only a static tree-table layout, i.e., there is no way to alter the
 * columns unless the developer changes the underlying {@link TreeTableView} directly. Note that if
 * a column is removed the index provided to the label provider might not be correct anymore.
 * <P>
 * Furthermore, the selection model is currently fixed to be row selection.
 */
public final class DynamicTreeTableViewer<T> extends DynamicTreeViewerBase<T> {
	/** The {@link TreeTableView} control to be managed. */
	private final TreeTableView<T> view;
	/** The UI provider of this tree-table. */
	private final DynamicTreeTableUIProviderBase<T> uiProvider;

	/**
	 * Constructor. The given {@link Supplier} is queried on updates to obtain a list of elements to
	 * display.
	 */
	public DynamicTreeTableViewer(TreeTableView<T> view, Supplier<Stream<T>> listFctn,
			DynamicStreamContentProvider<T, Supplier<Stream<T>>> contentProvider,
			DynamicTreeTableUIProviderBase<T> uiProvider) {
		this(view, new DynamicList<T>(listFctn, contentProvider), false, 0, contentProvider,
				uiProvider);
	}

	/**
	 * Constructor. The given input collection must be updated by the caller such that changes are
	 * reflected in the corresponding {@link TreeTableView}.
	 */
	public DynamicTreeTableViewer(TreeTableView<T> view, Collection<T> list,
			DynamicListContentProvider<T> contentProvider,
			DynamicTreeTableUIProviderBase<T> uiProvider) {
		this(view, new DynamicList<T>(list, contentProvider), false, 0, contentProvider,
				uiProvider);
	}

	/** Constructor. */
	public DynamicTreeTableViewer(TreeTableView<T> view, T root, boolean showRoot, int revealLevel,
			DynamicTreeContentProviderBase<T> contentProvider,
			DynamicTreeTableUIProviderBase<T> uiProvider) {
		this(view, new DynamicTreeItem<T>(root, contentProvider), showRoot, revealLevel,
				contentProvider, uiProvider);
	}

	/** Constructor. */
	public DynamicTreeTableViewer(T root, boolean showRoot, int revealLevel,
			DynamicTreeContentProviderBase<T> contentProvider,
			DynamicTreeTableUIProviderBase<T> uiProvider) {
		this(new TreeTableView<T>(), root, showRoot, revealLevel, contentProvider, uiProvider);
	}

	/** Constructor. */
	public DynamicTreeTableViewer(TreeTableView<T> view, DynamicTreeItemBase<T> rootItem,
			boolean showRoot, int revealLevel, DynamicTreeContentProviderBase<T> contentProvider,
			DynamicTreeTableUIProviderBase<T> uiProvider) {
		super(view, contentProvider);
		this.uiProvider = uiProvider;
		// construct view
		this.view = view;
		view.setRoot(rootItem);
		view.setShowRoot(showRoot);
		view.setEditable(true);
		view.setPlaceholder(new Label(""));
		view.setRowFactory(new Callback<TreeTableView<T>, TreeTableRow<T>>() {
			@Override
			public TreeTableRow<T> call(TreeTableView<T> param) {
				final TreeTableRow<T> row = new TreeTableRow<T>();
				row.setOnDragOver(evt -> {
					T item = row.getItem();
					if(uiProvider.dragOver(item, evt.getDragboard())) {
						evt.acceptTransferModes(TransferMode.ANY);
						evt.consume();
					}
				});
				row.setOnDragDropped(evt -> {
					T item = row.getItem();
					if(uiProvider.dropClipboardContent(item, evt.getDragboard())) {
						evt.setDropCompleted(true);
						evt.consume();
					}
				});
				return row;
			}
		});

		TreeTableViewSelectionModel<T> selectionModel = view.getSelectionModel();
		selectionModel.setSelectionMode(SelectionMode.MULTIPLE);
		selectionModel.selectedItemProperty().addListener((obs, oVal, nVal) -> {
			T ov = (oVal != null) ? oVal.getValue() : null;
			T nv = (nVal != null) ? nVal.getValue() : null;
			uiProvider.select(ov, nv);
		});

		rootItem.update();
		// expand to reveal (+1 if root node is not shown)
		expandItem(rootItem, showRoot ? revealLevel : revealLevel + 1);
	}

	/** Updates the viewer content. */
	public void update() {
		// wild cast works: see constructor
		DynamicTreeItemBase<T> rootItem = (DynamicTreeItemBase<T>)view.getRoot();

		List<T> selectedValues = getSelectedValues();
		view.getSelectionModel().clearSelection();
		rootItem.update();
		selectValues(rootItem, selectedValues);
		view.refresh();
	}

	/** Returns a list of all selected values */
	private List<T> getSelectedValues() {
		ObservableList<TreeItem<T>> selectedItems = view.getSelectionModel().getSelectedItems();
		return selectedItems.stream().map(c -> c.getValue()).collect(toList());
	}

	/** Get a stream of all items. */
	private Stream<TreeItem<T>> getAllItems() {
		return getItems(view.getRoot(), i -> false);
	}

	/** Get a stream of all visible items. */
	private Stream<TreeItem<T>> getVisibleItems() {
		return getItems(view.getRoot(), i -> !i.isExpanded());
	}

	/**
	 * Get a stream of items where sub-trees are pruned according to the given predicate.
	 * (This could be optimized to make stream generation lazy.)
	 */
	private Stream<TreeItem<T>> getItems(TreeItem<T> root, Predicate<TreeItem<T>> prune) {
		Stream<TreeItem<T>> items = Stream.of(root);
		if(!prune.test(root)) {
			for(TreeItem<T> child : root.getChildren()) {
				items = Stream.concat(items, getItems(child, prune));
			}
		}
		return items;
	}

	/** Scroll the tree view to make the first occurrence of the given value visible. */
	public void navigateTo(T value) {
		if(value == null) {
			return;
		}

		// Find value
		TreeItem<T> item = getAllItems() // all items
				.filter(i -> value.equals(i.getValue())) // all matching items
				.findFirst() // first matching item
				.orElse(null);
		if(item == null) {
			return;
		}

		// Ensure value is expanded
		TreeItem<T> parent = item.getParent();
		while(parent != null) {
			parent.setExpanded(true);
			parent = parent.getParent();
		}

		// Compute the count of the item
		int count = (int)getVisibleItems().takeWhile(i -> i != item).count();

		view.scrollTo(count);
	}

	/** Selects the given values in item's subtree */
	private void selectValues(TreeItem<T> item, List<T> values) {
		if(values.contains(item.getValue())) {
			view.getSelectionModel().select(item);
		}
		if(item.isExpanded()) {
			for(TreeItem<T> child : item.getChildren()) {
				selectValues(child, values);
			}
		}
	}

	/** Expands items up to the given level. */
	private void expandItem(TreeItem<T> parentItem, int revealLevel) {
		if(revealLevel <= 0) {
			return;
		}
		parentItem.setExpanded(true);
		revealLevel--;
		for(TreeItem<T> child : parentItem.getChildren()) {
			expandItem(child, revealLevel);
		}
	}

	/** Expand all items to the given level. */
	public void expandToLevel(int revealLevel) {
		TreeItem<T> root = view.getRoot();
		if(root == null) {
			return;
		}

		expandItem(root, revealLevel);
	}

	/** Expands all items on all levels. */
	public void expandAllItems() {
		expandToLevel(MAX_VALUE);
	}

	/**
	 * Adds a column to the table part of the view. The labels, context menus and icons are shown as
	 * defined in the {@link DynamicTreeTableUIProviderBase}.
	 */
	public TreeTableColumn<T, String> addColumn(String headerLabel, int prefWidth) {
		int num = view.getColumns().size();
		TreeTableColumn<T, String> column = new TreeTableColumn<>(headerLabel);
		column.setPrefWidth(prefWidth);

		column.setCellValueFactory(param -> {
			T data = param.getValue().getValue();
			SimpleObjectProperty<String> cell =
					new SimpleObjectProperty<String>(uiProvider.getLabel(data, num));
			return cell;
		});

		uiProvider.applyToColumn(num, column);

		view.getColumns().add(column);
		return column;
	}

	/**
	 * Adds an editable column to the table part of the view. The labels, context menus and icons
	 * are shown as defined in the {@link DynamicTreeTableUIProviderBase}.
	 */
	public TreeTableColumn<T, String> addTextColumn(String headerLabel, int prefWidth) {
		int num = view.getColumns().size();
		TreeTableColumn<T, String> column = addColumn(headerLabel, prefWidth);

		uiProvider.applyToTextColumn(num, column);

		return column;
	}

	/**
	 * Adds a boolean checkbox column to the table part of the view. The labels, context menus and
	 * icons are shown as defined in the {@link DynamicTreeTableUIProviderBase}.
	 * 
	 * <b>IMPORTANT</b>: the getLabel method of the uiProvider must return the boolean string "true"
	 * or "false" in order to work correctly
	 */
	public TreeTableColumn<T, Boolean> addCheckboxColumn(String headerLabel, int prefWidth) {
		int num = view.getColumns().size();
		TreeTableColumn<T, Boolean> column = new TreeTableColumn<>(headerLabel);
		column.setPrefWidth(prefWidth);
		column.setCellFactory(forTreeTableColumn(column));
		uiProvider.applyToCheckboxColumn(num, column);
		view.getColumns().add(column);
		return column;
	}

	/**
	 * Adds a column with Combo editing support to the view. The labels, context menus and
	 * icons are shown as defined in the {@link DynamicTreeTableUIProviderBase}.
	 * 
	 * @param headerLabel
	 *            The Label for the column header.
	 * @param prefWidth
	 *            The preferred width of the new column in px.
	 * @param comboValueFactory
	 *            A {@link Function} that maps cell elements to a combo value map (data model
	 *            element -> label)
	 * 
	 * @return The newly created an configured column.
	 */
	public TreeTableColumn<T, String> addComboColumn(String headerLabel, int prefWidth,
			Function<T, Map<?, String>> comboValueFactory) {

		int num = view.getColumns().size();
		TreeTableColumn<T, String> column = new TreeTableColumn<>(headerLabel);
		column.setPrefWidth(prefWidth);

		column.setCellValueFactory(param -> {
			T data = param.getValue().getValue();
			SimpleObjectProperty<String> cell =
					new SimpleObjectProperty<String>(uiProvider.getLabel(data, num));
			return cell;
		});

		uiProvider.applyToComboColumn(num, column, comboValueFactory);

		view.getColumns().add(column);

		return column;
	}

	/**
	 * Adds a column to the table part of the view. The labels, context menus and icons are shown as
	 * defined in the {@link DynamicTreeTableUIProviderBase}.
	 * 
	 * @deprecated use {@link DynamicTreeTableUIProviderBase}
	 */
	@Deprecated
	public TreeTableColumn<T, String> addColumn(String headerLabel, int prefWidth,
			@SuppressWarnings("unused") boolean readOnly) {
		return addColumn(headerLabel, prefWidth);
	}

	/** Returns the underlying {@link TreeTableView}. */
	public TreeTableView<T> getControl() {
		return view;
	}

	/** {@inheritDoc} */
	@Override
	protected T getSelectedObject() {
		final TreeItem<T> selectedItem = view.getSelectionModel().getSelectedItem();
		T value = selectedItem != null ? selectedItem.getValue() : null;

		return value;
	}
}
