/*--------------------------------------------------------------------------+
|                                                                          |
| Copyright 2021 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.variability.model;

import static java.util.stream.Collectors.toList;
import static org.fortiss.variability.util.VariabilityUtilsInternal.getAllElementsFromSameContainmentTree;

import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.variability.model.base.BaseFactory;
import org.fortiss.variability.model.base.RemoveBindingStrategy;
import org.fortiss.variability.model.features.AbstractAlternativeFeature;
import org.fortiss.variability.model.features.AbstractCompositionalFeature;
import org.fortiss.variability.model.features.AbstractFeature;
import org.fortiss.variability.model.features.AbstractFeatureModel;
import org.fortiss.variability.model.features.configuration.AlternativeFeatureConfiguration;
import org.fortiss.variability.model.features.configuration.CompositionalFeatureConfiguration;
import org.fortiss.variability.model.features.configuration.ConfigurationFactory;
import org.fortiss.variability.model.features.configuration.IFeatureConfiguration;
import org.fortiss.variability.model.features.configuration.VariantConfiguration;
import org.fortiss.variability.model.presence.AndPC;
import org.fortiss.variability.model.presence.DefaultPC;
import org.fortiss.variability.model.presence.ILiteralReferencable;
import org.fortiss.variability.model.presence.LiteralPC;
import org.fortiss.variability.model.presence.NotPC;
import org.fortiss.variability.model.presence.OrPC;
import org.fortiss.variability.model.presence.PresenceConditionTerm;
import org.fortiss.variability.model.presence.PresenceFactory;

/**
 * Model Element Factory for the variability plugin.

 * @author bayha
 *
 */
public class VariabilityModelElementFactory {

	/**
	 * Creates a new {@link OrPC}.
	 * 
	 * @param operand1 The left side of the disjunction.
	 * @param operand2 The right side of the disjunction.
	 * @param str      The String representation of this term.
	 * 
	 * @return The newly created term.
	 */
	public static OrPC createOrPC(PresenceConditionTerm operand1, PresenceConditionTerm operand2, String str) {
		OrPC or = PresenceFactory.eINSTANCE.createOrPC();

		or.setStringRepresentation(str);
		or.setOperand1(operand1);
		or.setOperand2(operand2);

		return or;
	}

	/**
	 * Creates a new {@link AndPC}.
	 * 
	 * @param operand1 The left side of the conjunction.
	 * @param operand2 The right side of the conjunction.
	 * @param str      The String representation of this term.
	 * 
	 * @return The newly created term.
	 */
	public static AndPC createAndPC(PresenceConditionTerm operand1, PresenceConditionTerm operand2, String str) {
		AndPC and = PresenceFactory.eINSTANCE.createAndPC();

		and.setStringRepresentation(str);
		and.setOperand1(operand1);
		and.setOperand2(operand2);

		return and;
	}

	/**
	 * Creates a new {@link NotPC}.
	 * 
	 * @param operand The term to be negated.
	 * @param str     The String representation of this term.
	 * 
	 * @return The newly created negated term.
	 */
	public static NotPC createNotPC(PresenceConditionTerm operand, String str) {
		NotPC not = PresenceFactory.eINSTANCE.createNotPC();

		not.setStringRepresentation(str);
		not.setOperand(operand);

		return not;
	}

	/**
	 * Creates a new {@link LiteralPC}.
	 * 
	 * @param literalString The name of the literal.
	 * @param context       Any element, that is contained in the same containment
	 *                      tree as the element, this literal shall point to.
	 * 
	 * @return The newly created {@link LiteralPC}.
	 */
	public static LiteralPC createLiteralPC(String literalString, EObject context) {
		LiteralPC ret = PresenceFactory.eINSTANCE.createLiteralPC();
		ret.setStringRepresentation(literalString);

		List<ILiteralReferencable> literalRefs = getAllElementsFromSameContainmentTree(context,
				ILiteralReferencable.class);

		for (ILiteralReferencable l : literalRefs) {
			if (l.getName().equals(literalString)) {
				ret.setLiteralReference(l);

				break;
			}
		}
		return ret;
	}

	/**
	 * Creates a new {@link LiteralPC}.
	 * 
	 * @param literalRef The element, this literal shall point to.
	 * 
	 * @return The newly created {@link LiteralPC}.
	 */
	public static LiteralPC createLiteralPC(ILiteralReferencable literalRef) {
		LiteralPC ret = PresenceFactory.eINSTANCE.createLiteralPC();

		ret.setLiteralReference(literalRef);
		ret.setStringRepresentation(literalRef.getName());

		return ret;
	}

	/**
	 * Creates new {@link DefaultPC}.
	 * 
	 * @return The newly created {@link DefaultPC}.
	 */
	public static DefaultPC createDefaultPC() {
		DefaultPC def = PresenceFactory.eINSTANCE.createDefaultPC();

		def.setStringRepresentation("DEFAULT");

		return def;
	}

	/**
	 * Creates a new, empty {@link VariantConfiguration}, to configure variants
	 * according to the given {@link AbstractFeatureModel}.
	 * 
	 * @param featureModel The {@link AbstractFeatureModel} according to which
	 *                     variants are supposed to be configured.
	 * 
	 * @return The newly created {@link VariantConfiguration}.
	 */
	public static VariantConfiguration createVariantConfiguration(AbstractFeatureModel featureModel) {
		VariantConfiguration varConfig = ConfigurationFactory.eINSTANCE.createVariantConfiguration();
		varConfig.setSelected(true);
		varConfig.setFeatureModel(featureModel);

		List<IFeatureConfiguration<?>> subConfigs = featureModel.getSubFeatures().stream()
				.map(f -> createFeatureConfigurationForFeature(f)).collect(toList());
		varConfig.getSubConfigurations().addAll(subConfigs);

		return varConfig;
	}

	/**
	 * Recursively creates a {@link IFeatureConfiguration} for the given
	 * {@link AbstractFeature} and its sub features.
	 * 
	 * @param feature The {@link AbstractFeature} that shall be configured.
	 * 
	 * @return The newly created {@link IFeatureConfiguration}.
	 */
	public static IFeatureConfiguration<?> createFeatureConfigurationForFeature(AbstractFeature feature) {
		if (feature instanceof AbstractAlternativeFeature) {
			AlternativeFeatureConfiguration afc = ConfigurationFactory.eINSTANCE
					.createAlternativeFeatureConfiguration();
			afc.setFeatureReference((AbstractAlternativeFeature) feature);

			if (feature.isOptional()) {
				afc.setSelected(false);
			} else {
				afc.setSelected(true);
			}

			return afc;
		}
		if (feature instanceof AbstractCompositionalFeature) {
			CompositionalFeatureConfiguration cfc = ConfigurationFactory.eINSTANCE
					.createCompositionalFeatureConfiguration();
			cfc.setFeatureReference((AbstractCompositionalFeature) feature);

			if (feature.isOptional()) {
				cfc.setSelected(false);
			} else {
				cfc.setSelected(true);

				EList<IFeatureConfiguration<?>> subConfigurations = cfc.getSubConfigurations();
				AbstractCompositionalFeature acf = (AbstractCompositionalFeature) feature;
				for (AbstractFeature sf : acf.getSubFeatures()) {
					subConfigurations.add(createFeatureConfigurationForFeature(sf));
				}
			}

			return cfc;
		}
		return null;
	}

	/** Singleton for {@link RemoveBindingStrategy}. */
	private static final RemoveBindingStrategy REMOVE_BINDING_STRATEGY_SINGLETON = BaseFactory.eINSTANCE
			.createRemoveBindingStrategy();

	/**
	 * Retrieves the singleton {@link RemoveBindingStrategy}.
	 * 
	 * @return {@link RemoveBindingStrategy} singleton.
	 */
	public static RemoveBindingStrategy getRemoveBindingStrategyInstance() {
		return REMOVE_BINDING_STRATEGY_SINGLETON;
	}
}
