/*--------------------------------------------------------------------------+
 |                                                                          |
 | Copyright 2021 fortiss GmbH                   						    |
 |                                                                          |
 | Licensed under the Apache License, Version 2.0 (the "License");          |
 | you may not use this file except in compliance with the License.         |
 | You may obtain a copy of the License at                                  |
 |                                                                          |
 |    http://www.apache.org/licenses/LICENSE-2.0                            |
 |                                                                          |
 | Unless required by applicable law or agreed to in writing, software      |
 | distributed under the License is distributed on an "AS IS" BASIS,        |
 | WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
 | See the License for the specific language governing permissions and      |
 | limitations under the License.                                           |
 +--------------------------------------------------------------------------*/
package org.fortiss.variability.util;

import static org.eclipse.emf.ecore.util.EcoreUtil.getAllContents;
import static org.eclipse.emf.ecore.util.EcoreUtil.getRootContainer;
import static org.eclipse.emf.ecore.util.EcoreUtil.UsageCrossReferencer.find;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.function.Predicate;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.util.EcoreUtil;

/**
 * Utility functions that are used within the variability layers.
 * 
 * @author bayha
 */
public class VariabilityUtilsInternal {

	/**
	 * Retrieves all elements of type T from a containment tree. The 'context' is
	 * any element in a containment tree. The result will also contain all parents,
	 * siblings and children with the specified type.
	 * 
	 * @param context
	 *            Any object, contained in the containment tree to look in.
	 * @param clazz
	 *            The {@link Class} of the elements to look for.
	 * @return A {@link List} of all elements of type T from the containment tree.
	 */
	public static <T extends EObject> List<T> getAllElementsFromSameContainmentTree(EObject context,
			Class<T> clazz) {

		if(context == null) {
			throw new IllegalArgumentException(
					"The given context is either null or is not contained in a FileProject!");
		}

		// Find topmost EObject from the containment tree.
		EObject root = context;
		while(root.eContainer() != null) {
			root = root.eContainer();
		}

		return getChildrenWithType(root, clazz);
	}

	/**
	 * Returns the value from the map, whose key is the best match for the given
	 * class. The best match is defined by the first match occurring in a breath
	 * first search of the inheritance tree, where the base class is always
	 * visited before the implemented interfaces. Interfaces are traversed in
	 * the order they are defined in the source file. The only exception is
	 * {@link Object}, which is considered only as the very last option.
	 * <p>
	 * As this lookup can be expensive (reflective iteration over the entire
	 * inheritance tree) the results should be cached if multiple lookups for
	 * the same class are expected.
	 * 
	 * 
	 * @param clazz
	 *            the class being looked up.
	 * @param classMap
	 *            the map to perform the lookup in.
	 * @return the best match found or <code>null</code> if no matching entry
	 *         was found. Note that <code>null</code> will also be returned if
	 *         the entry for the best matching class was <code>null</code>.
	 */
	public static <T> T performNearestClassLookup(Class<?> clazz, Map<Class<?>, T> classMap) {
		Queue<Class<?>> q = new LinkedList<Class<?>>();
		q.add(clazz);

		while(!q.isEmpty()) {
			Class<?> current = q.poll();
			if(classMap.containsKey(current)) {
				return classMap.get(current);
			}

			Class<?> superClass = current.getSuperclass();
			if(superClass != null && superClass != Object.class) {
				q.add(superClass);
			}

			for(Class<?> iface : current.getInterfaces()) {
				q.add(iface);
			}
		}
		return classMap.get(Object.class);
	}

	/**
	 * For a given {@link EObject} recursively returns all its children that satisfy the given
	 * {@link Predicate}.
	 * 
	 * @param parent
	 *            the parent {@link EObject}.
	 * @param pred
	 *            {@link Predicate} to be satisfied by the children.
	 * @return a list of children.
	 */
	@SuppressWarnings("unchecked")
	public static <S> EList<S> getChildrenWith(EObject parent, Predicate<EObject> pred) {
		EList<S> children = new BasicEList<S>();
		TreeIterator<EObject> content = getAllContents(parent, true);

		while(content.hasNext()) {
			EObject child = content.next();
			if(pred.test(child)) {
				children.add((S)child);
			}
		}
		return children;
	}

	/**
	 * For a given {@link EObject} recursively returns all its children that have the given type.
	 * 
	 * @param parent
	 *            the parent {@link EObject}.
	 * @param type
	 *            the type.
	 * @return a list of children.
	 */
	public static <S> EList<S> getChildrenWithType(EObject parent, Class<S> type) {
		Predicate<EObject> isAssignable = element -> type.isAssignableFrom(element.getClass());
		return getChildrenWith(parent, isAssignable);
	}

	/**
	 * Returns a list of all parent {@link EObject}s that satisfy the given {@link Predicate}. If
	 * none
	 * are found, an empty list is returned.
	 * 
	 * @param child
	 *            The {@link EObject} from which the "upwards" search shall begin.
	 * @param pred
	 *            {@link Predicate} to be satisfied by parent elements.
	 * @return List of parent {@link EObject}s of type <code>type</code>.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends EObject> EList<T> getParentsWith(EObject child,
			Predicate<EObject> pred) {

		EList<T> parentElements = new BasicEList<T>();
		EObject currentParent = child.eContainer();

		while(currentParent != null) {
			if(pred.test(currentParent)) {
				parentElements.add((T)currentParent);
			}
			currentParent = currentParent.eContainer();
		}
		return parentElements;
	}

	/**
	 * Returns a list of all parent {@link EObject}s that have the type <code>type</code>. If none
	 * are found, an empty list is returned.
	 * 
	 * @param child
	 *            The {@link EObject} from which the "upwards" search shall begin.
	 * @param type
	 *            The class type which is used to filter the search.
	 * @return List of parent {@link EObject}s of type <code>type</code>.
	 */
	public static <T extends EObject> EList<T> getParentsWithType(EObject child, Class<T> type) {
		Predicate<EObject> isAssignable = element -> type.isAssignableFrom(element.getClass());
		return getParentsWith(child, isAssignable);
	}

	/**
	 * Retrieves all EStructuralFeature.Setting, which are a reference to the given
	 * 'referencedEObject'.
	 * 
	 * Note: This implementation is adopted from {@link EcoreUtil}.delete().
	 * 
	 * @param referencedEObject
	 *            The {@link EObject}, to which all references are supposed to be retrieved.
	 * @return A {@link Collection}, containing all EStructuralFeature.Setting>, which are
	 *         references
	 *         to 'referencedEObject'.
	 */
	public static Collection<EStructuralFeature.Setting>
			getAllReferences(EObject referencedEObject) {
		EObject rootEObject = getRootContainer(referencedEObject);
		Resource resource = rootEObject.eResource();

		Collection<EStructuralFeature.Setting> usages;
		if(resource == null) {
			usages = find(referencedEObject, rootEObject);
		} else {
			ResourceSet resourceSet = resource.getResourceSet();
			usages = (resourceSet == null) ? find(referencedEObject, resource)
					: find(referencedEObject, resourceSet);
		}
		return usages;
	}

}
